@@ -1,7 +1,7 @@
 /* helpers.h - Various helper functions
-   Copyright (C) 2000, 2001 Thomas Moestl
 
-   With modifications by Paul Rombouts, 2002, 2003, 2004.
+   Copyright (C) 2000, 2001 Thomas Moestl
+   Copyright (C) 2002, 2003, 2004 Paul A. Rombouts
 
 This file is part of the pdnsd package.
 
@@ -35,10 +35,11 @@ Boston, MA 02111-1307, USA.  */
 
 #define SOFTLOCK_MAXTRIES 1000
 
-int run_as(char *user);
+int run_as(const char *user);
 void pdnsd_exit(void);
 int softlock_mutex(pthread_mutex_t *mutex);
 
+#if 0
 inline static int isdchar (unsigned char c)
 {
   return ((c>='a' && c<='z') || (c>='A' && c<='Z') || (c>='0' && c<='9') || c=='-'
@@ -47,8 +48,9 @@ inline static int isdchar (unsigned char c)
 #endif
 	  );
 }
+#endif
 
-void rhn2str(const unsigned char *rhn, unsigned char *str);
+const unsigned char *rhn2str(const unsigned char *rhn, unsigned char *str, int size);
 int  str2rhn(const unsigned char *str, unsigned char *rhn);
 const char *parsestr2rhn(const unsigned char *str, int len, unsigned char *rhn);
 
@@ -58,19 +60,41 @@ const char *parsestr2rhn(const unsigned char *str, int len, unsigned char *rhn);
 */
 inline static unsigned int rhnlen(const unsigned char *rhn)
 {
-	unsigned int i=0;
-	unsigned char lb;
+	unsigned int i=0,lb;
 
 	while((lb=rhn[i]))
 		i+=lb+1;
 	return i+1;
 }
 
-/* count the number of name segments. */
+/* Skip k segments in a name in length-byte string notation. */
+inline static unsigned char *skipsegs(unsigned char *nm, unsigned k)
+{
+	unsigned lb;
+	for(;k;--k) {
+		lb= *nm;
+		if(!lb) return nm;
+		nm += lb+1;
+	}
+	return nm;
+}
+
+/* Skip a name in length-byte string notation and return a pointer to the
+   position right after the terminating null byte.
+*/
+inline static unsigned char *skiprhn(unsigned char *rhn)
+{
+	unsigned lb;
+
+	while((lb= *rhn))
+		rhn += lb+1;
+	return rhn+1;
+}
+
+/* count the number of name segments of a name in length-byte string notation. */
 inline static unsigned int rhnsegcnt(const unsigned char *rhn)
 {
-	unsigned int res=0;
-	unsigned char lb;
+	unsigned int res=0,lb;
 
 	while((lb= *rhn)) {
 		++res;
@@ -81,7 +105,7 @@ inline static unsigned int rhnsegcnt(const unsigned char *rhn)
 
 unsigned int rhncpy(unsigned char *dst, const unsigned char *src);
 
-int follow_cname_chain(dns_cent_t *c, unsigned char *name, unsigned char *rrn);
+int follow_cname_chain(dns_cent_t *c, unsigned char *name);
 
 inline static int is_inaddr_any(pdnsd_a *a)
 {
@@ -130,10 +154,11 @@ extern FILE *rand_file;
 unsigned short get_rand16(void);
 
 int fsprintf(int fd, const char *format, ...) printfunc(2, 3);
-#if defined(__GNUC__) && (__GNUC__ < 2 || (__GNUC__ == 2 && __GNUC_MINOR__ < 95))
+#if !defined(CPP_C99_VARIADIC_MACROS)
+/* GNU C Macro Varargs style. */
 # define fsprintf_or_return(args...) {int _retval; if((_retval=fsprintf(args))<0) return _retval;}
 #else
-/* ANSI style variadic macro. */
+/* ANSI C99 style variadic macro. */
 # define fsprintf_or_return(...) {int _retval; if((_retval=fsprintf(__VA_ARGS__))<0) return _retval;}
 #endif
 
@@ -154,11 +179,15 @@ inline static int write_all(int fd,const void *data,int n)
   return written;
 }
 
+void hexdump(const void *data, int dlen, char *buf, int buflen);
+int escapestr(char *in, int ilen, char *str, int size);
 
+#if 0
 inline static int stricomp(const char *a, const char *b)
 {
   return !strcasecmp(a,b);
 }
+#endif
 
 /* compare two names in length byte - string format */
 inline static int rhnicmp(const unsigned char *a, const unsigned char *b)
@@ -170,17 +199,20 @@ inline static int rhnicmp(const unsigned char *a, const unsigned char *b)
 		if(lb!=b[i]) return 0;
 		if(!lb) break;
 		++i;
-		for(;lb;--lb) {
+		do {
 			if(tolower(a[i])!=tolower(b[i])) return 0;
 			++i;
-		}
+		} while(--lb);
 	}
 	return 1;
 }
 
 /* Bah. I want strlcpy. */
-inline static int strncp(char *dst, const char *src, int dstsz)
+inline static int strncp(char *dst, const char *src, size_t dstsz)
 {
+#ifdef HAVE_STRLCPY
+	return (strlcpy(dst,src,dstsz)<dstsz);
+#else
 #ifdef HAVE_STPNCPY
 	char *p=stpncpy(dst,src,dstsz);
 	if(p<dst+dstsz) return 1;
@@ -192,6 +224,7 @@ inline static int strncp(char *dst, const char *src, int dstsz)
 	dst[dstsz-1]='\0';
 	return 0;
 #endif
+#endif
 }
 
 #ifndef HAVE_STRDUP
@@ -233,6 +266,14 @@ inline static char *stpcpy (char *dest, const char *src)
 }
 #endif
 
+#ifndef HAVE_MEMPCPY
+inline static void *mempcpy(void *dest, const void *src, size_t len)
+{
+  memcpy(dest,src,len);
+  return ((char *)dest)+len;
+}
+#endif
+
 #ifndef HAVE_GETLINE
 int getline(char **lineptr, size_t *n, FILE *stream);
 #endif
@@ -241,6 +282,15 @@ int getline(char **lineptr, size_t *n, FILE *stream);
 int asprintf (char **lineptr, const char *format, ...);
 #endif
 
+#ifndef HAVE_VASPRINTF
+#include <stdarg.h>
+int vasprintf (char **lineptr, const char *format, va_list va);
+#endif
+
+#ifndef HAVE_INET_NTOP
+const char *inet_ntop(int af, const void *src, char *dst, size_t size);
+#endif
+
 #define strlitlen(strlit) (sizeof(strlit)-1)
 
 #endif /* HELPERS_H */
